package org.briarproject.bramble.plugin.lora;

import org.briarproject.bramble.plugin.lora.AbstractLoRaPlugin;
import org.briarproject.bramble.api.plugin.TransportConnectionWriter;
import java.io.IOException;
import java.io.OutputStream;
import org.briarproject.bramble.plugin.lora.LoRaAdapter;
import java.util.Arrays;
import java.lang.System;
import java.util.logging.Logger;
import org.briarproject.bramble.util.ByteUtils;
import org.briarproject.bramble.plugin.lora.L2PTagManager;
import org.briarproject.bramble.api.Bytes;

import static org.briarproject.bramble.api.plugin.LoRaConstants.*;
import static org.briarproject.bramble.api.transport.TransportConstants.TAG_LENGTH;

public class L2PWriter implements TransportConnectionWriter {

	private static final Logger LOG =
			Logger.getLogger(L2PWriter.class.getName());

    private AbstractLoRaPlugin plugin;

    L2PWriter(AbstractLoRaPlugin plugin) {
		LOG.info("L2PWriter::Constructor");
        this.plugin = plugin;
    }

    @Override
    public long getMaxLatency() {
        return plugin.getMaxLatency();
    }

    @Override
    public int getMaxIdleTime() {
        return plugin.getMaxIdleTime();
    }

    @Override
    public boolean isLossyAndCheap() {
        return false;
    }

    @Override
    public OutputStream getOutputStream() throws IOException {
        return new L2OutputStream(plugin.getAdapter());
    }

    @Override
    public void dispose(boolean exception) throws IOException {
    }

    private class L2OutputStream extends OutputStream {
        private LoRaAdapter adapter;
        private byte[] buffer;
        private byte[] btpTag = new byte[TAG_LENGTH];
        private short readTagLength=0;
        private long sequenceNr=0;

        L2OutputStream(LoRaAdapter adapter) {
            this.adapter = adapter;
            buffer = new byte[0];
        }
    
        @Override
        public void close() throws IOException {
            flush();
            LOG.info("close()");
        }
    
        @Override
        public void write(int b) throws IOException {
            write(new byte[]{(byte)b}, 0, 1);
        }
    
        @Override
        public void write(byte[] b) throws IOException {
            write(b, 0, b.length);
            //flush();
        }

        @Override
        public void write(byte[] b, int off, int len) throws IOException {
            LOG.info("WRITE ("+len+" Bytes)");
            LOG.info(ByteUtils.bytesToHex(b));

            //Read BTP-Tag first (if not done) and write it in another Array
            while(readTagLength < TAG_LENGTH && len > 0) {
                btpTag[readTagLength++] = b[off++];
                len--;
            }

            if(len <= 0) //for speed
                return;

            if(off != 0) {
                byte tmp[] = new byte[len];
                System.arraycopy(b,off,tmp,0,len);
                b = tmp;
            } else if(b.length != len)
                b = Arrays.copyOf(b,len);

            buffer = ByteUtils.concat(buffer, b);
            while(buffer.length >= PAYLOAD_SIZE) {
                flush();
            }
        }

        @Override
        public void flush() {
            if(buffer.length != 0) {
                int len = Math.min(PAYLOAD_SIZE, buffer.length);

                byte[] pkg = new byte[len+TAG_LENGTH];
                System.arraycopy(L2PTagManager.genL2PTag(new Bytes(btpTag), sequenceNr++).getBytes(), 0, pkg, 0, TAG_LENGTH); //Write L2PTag
                System.arraycopy(buffer, 0, pkg, TAG_LENGTH, len);
                adapter.send(pkg);
                
                buffer = Arrays.copyOfRange(buffer, len, buffer.length);
            }
        }
    }
}